package de.moritzruth.dracula_musical.device

import de.moritzruth.theaterdsl.device.AngleDV
import de.moritzruth.theaterdsl.device.ConcreteDV
import de.moritzruth.theaterdsl.device.Device
import de.moritzruth.theaterdsl.device.PercentageDV
import de.moritzruth.theaterdsl.dmx.DmxAddress
import de.moritzruth.theaterdsl.dmx.DmxDataWriter
import de.moritzruth.theaterdsl.dmx.DmxValue
import de.moritzruth.theaterdsl.value.Percentage
import de.moritzruth.theaterdsl.value.degrees
import de.moritzruth.theaterdsl.value.percent
import kotlinx.collections.immutable.persistentSetOf
import kotlin.math.roundToInt
import kotlin.time.Duration

class CoemarProWash(override val firstChannel: DmxAddress, private val invertAngles: Boolean) : Device {
    companion object {
        private const val PAN_LIMIT = 530f // degrees
        private const val TILT_LIMIT = 284f // degrees
    }

    override val numberOfChannels: UInt = 16u

    sealed interface ColorWheelMode {
        fun getDmxValue(): DmxValue
        sealed class Simple(staticDmxValue: Int) : ColorWheelMode {
            val staticDmxValue = DmxValue(staticDmxValue.toUByte())
            override fun getDmxValue() = staticDmxValue
        }

        class Rotate(val speed: Percentage, val backwards: Boolean = false) : ColorWheelMode {
            @Suppress("SimplifyBooleanWithConstants", "KotlinConstantConditions")
            override fun getDmxValue() = run {
                if (backwards || true) ((1.0 - speed.value) * 18).roundToInt() + 168
                else ((1.0 - speed.value) * 18).roundToInt() + 128 // TODO: Fix
            }.let { DmxValue(it.toUByte()) }
        }

        object White : Simple(0)
        object Red : Simple(26)
        object DarkBlue : Simple(44)
        object Green : Simple(62)
        object WarmWhite : Simple(81)
        object Violet : Simple(118)
    }

    val pan = AngleDV()
    val tilt = AngleDV()
    val brightness = PercentageDV()
    val beamAngle = PercentageDV(100.percent)
    val strobeSpeed = PercentageDV(0.percent)

    val colorWheelMode = ConcreteDV<ColorWheelMode>(ColorWheelMode.White)
    val cyan = PercentageDV()
    val magenta = PercentageDV()
    val yellow = PercentageDV()

    fun pointAtStageCenter() {
        tilt.static(55.degrees)
        pan.static(if (invertAngles) 280.degrees else 80.degrees)
    }

    fun pointAtCeiling(fadeDuration: Duration = Duration.ZERO) {
        pan.fade(360.degrees, fadeDuration)
        tilt.fade(120.degrees, fadeDuration)
    }

    fun pointAtAudience(fadeDuration: Duration = Duration.ZERO) {
        pan.fade(360.degrees, fadeDuration)
        tilt.fade(240.degrees, fadeDuration)
    }

    override val dvs = persistentSetOf(
        pan, tilt, brightness, beamAngle, strobeSpeed, colorWheelMode, cyan, magenta, yellow
    )

    override fun writeDmxData(writer: DmxDataWriter, isLightBehindCurtainOn: Boolean) {
        val pan = (pan.getCurrentValue().degrees.mod(PAN_LIMIT) / PAN_LIMIT * 65536).roundToInt().toUShort()
        val tilt = (tilt.getCurrentValue().degrees.mod(TILT_LIMIT) / TILT_LIMIT * 65536).roundToInt().toUShort()

        writer.writeHighByte(pan)
        writer.writeHighByte(tilt)
        writer.writeLowByte(pan)
        writer.writeLowByte(tilt)
        writer.writeRaw(DmxValue(0u)) // pan/tilt speed = maximum
        writer.writeRaw(DmxValue(255u)) // fan speed and lamp control = no function
        writer.writeRaw(colorWheelMode.getCurrentValue().getDmxValue())
        writer.writePercentage(cyan.getCurrentValue())
        writer.writePercentage(magenta.getCurrentValue())
        writer.writePercentage(yellow.getCurrentValue())
        writer.writeRaw(DmxValue(0u)) // dimmer and color speed = maximum
        writer.writeRaw(DmxValue(0u)) // colour macro = off

//        writer.writeRaw(DmxValue((beamAngle.getCurrentValue().value * 54 + 125).roundToInt().toUByte()))
        writer.writeRaw(DmxValue(0u))

        writer.writeRaw(DmxValue(0u)) // no function
        writer.writeRaw(DmxValue((strobeSpeed.getCurrentValue().value * 32 + 63).roundToInt().toUByte()))
        writer.writePercentage(brightness.getCurrentValue())
    }
}